#include <cuda.h>
#include <cmath>
#include <iostream>
#include "MersenneTwister.h"
using namespace std;

MTRand rand1;
const int monte_carlo_steps = 50000; //For simplicity, should be a multiple of 1000
__global__ void blairsFerry(double* results, double* rands)
{	
	//3.6 miles @ 35 mph; 8 lights: 50% chance per hitting; 2 min. delay/light
	double lightSpacing = (3.6/8);
	double speed = (35.0/60); //Miles per minute
	double timeBtwLight = lightSpacing / speed; //Minutes btw. lights
	
	double timeElapsed = 0; //In minutes
	int lightsPassed = 0;
	while(lightsPassed < 8)
	{
		timeElapsed += timeBtwLight;
		double rand = rands[blockIdx.x*1000 + threadIdx.x*8 + lightsPassed];
		if(rand < 0.5) //Light is hit
			timeElapsed+=2;
	
		lightsPassed++;
	}
	results[blockIdx.x*1000+threadIdx.x] = timeElapsed;
}

__global__ void collins(double* results, double* rands)
{
	double lightSpacing = (2.2/7);
	double speed = (25.0/60);
	double timeBtwLight = lightSpacing/speed;
	double timeElapsed = 0;
	int lightsPassed = 0;

	timeElapsed += 0.7/speed; //Initial, lightless section
	while(lightsPassed < 7)
	{
		timeElapsed += timeBtwLight;
		double rand = rands[blockIdx.x*1000 + threadIdx.x*7 + lightsPassed];
		if(rand < 0.35) //Light is hit
			timeElapsed+=3;

		lightsPassed++;
	}

	results[blockIdx.x*1000+threadIdx.x] = timeElapsed;
}

double monteCarlo(char q)
{
	double* results = (double*)calloc(monte_carlo_steps, sizeof(double));
	double* cudaResults;
	cudaMalloc(&cudaResults, monte_carlo_steps*sizeof(double));
	cudaMemcpy(cudaResults, results, monte_carlo_steps*sizeof(double), cudaMemcpyHostToDevice);

	double* rands = (double*)malloc(500*sizeof(double)*monte_carlo_steps);
	double* cudaRands;
	cudaMalloc(&cudaRands, 500*sizeof(double)*monte_carlo_steps);
	for(int x = 0; x < 500*monte_carlo_steps; x++)
		rands[x] = rand1.rand(1);
	cudaMemcpy(cudaRands, rands, 500*sizeof(double)*monte_carlo_steps, cudaMemcpyHostToDevice);
	double R = 0.0;
	switch(q)
	{
	case 'c':	collins<<<monte_carlo_steps/1000,1000>>>(cudaResults, cudaRands); 
				cudaMemcpy(results, cudaResults, sizeof(double)*monte_carlo_steps, cudaMemcpyDeviceToHost);	
				for(int i = 0; i < monte_carlo_steps; i++)			
					R += results[i];

				R/=monte_carlo_steps;
				free(results);
				cudaFree(cudaResults);
				cudaFree(cudaRands);
				return R;
	 
	case 'b':   blairsFerry<<<monte_carlo_steps/1000,1000>>>(cudaResults, cudaRands);
				cudaMemcpy(results, cudaResults, sizeof(double)*monte_carlo_steps, cudaMemcpyDeviceToHost);
				double R = 0.0;		
				for(int i = 0; i < monte_carlo_steps; i++)
					R += results[i];
				R/=monte_carlo_steps;				
				free(results);
				cudaFree(cudaResults);
				cudaFree(cudaRands);
				return R;
	}
	free(results);
	cudaFree(cudaResults);
	cudaFree(cudaRands);
	return 0.0;
}
int main()
{
	double collins = monteCarlo('c');
	double blairs = monteCarlo('b');
	cout << "Collins Results: " << collins << endl;
	cout << "Collins ERROR: " << collins-14.31 << "(" << abs(((collins-14.31)/collins)*100) << "%)" << endl;
	cout << "Blair's Ferry Results: " << blairs << endl;
	cout << "Blair's Ferry ERROR: " << blairs-14.17 << "(" << abs(((blairs-14.17)/blairs)*100) << "%)" << endl;
	system("PAUSE");
	return 0;
}